@@ -19,18 +19,22 @@ import (
 	"io/ioutil"
 	"os"
 	"sort"
-	"strings"
 	"testing"
 	"time"
 
 	"github.com/go-kit/kit/log"
-
 	"github.com/prometheus/prometheus/pkg/labels"
+	"github.com/prometheus/prometheus/pkg/timestamp"
 	"github.com/prometheus/prometheus/promql/parser"
 	"github.com/prometheus/prometheus/storage"
 	"github.com/prometheus/prometheus/util/testutil"
+	"go.uber.org/goleak"
 )
 
+func TestMain(m *testing.M) {
+	goleak.VerifyTestMain(m)
+}
+
 func TestQueryConcurrency(t *testing.T) {
 	maxConcurrency := 10
 
@@ -117,9 +121,7 @@ func TestQueryTimeout(t *testing.T) {
 	testutil.NotOk(t, res.Err, "expected timeout error but got none")
 
 	var e ErrQueryTimeout
-	// TODO: when circleci-windows moves to go 1.13:
-	// testutil.Assert(t, errors.As(res.Err, &e), "expected timeout error but got: %s", res.Err)
-	testutil.Assert(t, strings.HasPrefix(res.Err.Error(), e.Error()), "expected timeout error but got: %s", res.Err)
+	testutil.Assert(t, errors.As(res.Err, &e), "expected timeout error but got: %s", res.Err)
 }
 
 const errQueryCanceled = ErrQueryCanceled("test statement execution")
@@ -175,8 +177,8 @@ type errQuerier struct {
 	err error
 }
 
-func (q *errQuerier) Select(bool, *storage.SelectHints, ...*labels.Matcher) (storage.SeriesSet, storage.Warnings, error) {
-	return errSeriesSet{err: q.err}, nil, q.err
+func (q *errQuerier) Select(bool, *storage.SelectHints, ...*labels.Matcher) storage.SeriesSet {
+	return errSeriesSet{err: q.err}
 }
 func (*errQuerier) LabelValues(string) ([]string, storage.Warnings, error) { return nil, nil, nil }
 func (*errQuerier) LabelNames() ([]string, storage.Warnings, error)        { return nil, nil, nil }
@@ -187,9 +189,10 @@ type errSeriesSet struct {
 	err error
 }
 
-func (errSeriesSet) Next() bool         { return false }
-func (errSeriesSet) At() storage.Series { return nil }
-func (e errSeriesSet) Err() error       { return e.err }
+func (errSeriesSet) Next() bool                   { return false }
+func (errSeriesSet) At() storage.Series           { return nil }
+func (e errSeriesSet) Err() error                 { return e.err }
+func (e errSeriesSet) Warnings() storage.Warnings { return nil }
 
 func TestQueryError(t *testing.T) {
 	opts := EngineOpts{
@@ -211,46 +214,36 @@ func TestQueryError(t *testing.T) {
 
 	res := vectorQuery.Exec(ctx)
 	testutil.NotOk(t, res.Err, "expected error on failed select but got none")
-	testutil.Equals(t, errStorage, res.Err)
+	testutil.Assert(t, errors.Is(res.Err, errStorage), "expected error doesn't match")
 
 	matrixQuery, err := engine.NewInstantQuery(queryable, "foo[1m]", time.Unix(1, 0))
 	testutil.Ok(t, err)
 
 	res = matrixQuery.Exec(ctx)
 	testutil.NotOk(t, res.Err, "expected error on failed select but got none")
-	testutil.Equals(t, errStorage, res.Err)
+	testutil.Assert(t, errors.Is(res.Err, errStorage), "expected error doesn't match")
 }
 
-// hintCheckerQuerier implements storage.Querier which checks the start and end times
-// in hints.
-type hintCheckerQuerier struct {
-	start    int64
-	end      int64
-	grouping []string
-	by       bool
-	selRange int64
-	function string
-
-	t *testing.T
+type noopHintRecordingQueryable struct {
+	hints []*storage.SelectHints
 }
 
-func (q *hintCheckerQuerier) Select(_ bool, sp *storage.SelectHints, _ ...*labels.Matcher) (storage.SeriesSet, storage.Warnings, error) {
-	testutil.Equals(q.t, q.start, sp.Start)
-	testutil.Equals(q.t, q.end, sp.End)
-	testutil.Equals(q.t, q.grouping, sp.Grouping)
-	testutil.Equals(q.t, q.by, sp.By)
-	testutil.Equals(q.t, q.selRange, sp.Range)
-	testutil.Equals(q.t, q.function, sp.Func)
-
-	return errSeriesSet{err: nil}, nil, nil
+func (h *noopHintRecordingQueryable) Querier(context.Context, int64, int64) (storage.Querier, error) {
+	return &hintRecordingQuerier{Querier: &errQuerier{}, h: h}, nil
 }
-func (*hintCheckerQuerier) LabelValues(string) ([]string, storage.Warnings, error) {
-	return nil, nil, nil
-}
-func (*hintCheckerQuerier) LabelNames() ([]string, storage.Warnings, error) { return nil, nil, nil }
-func (*hintCheckerQuerier) Close() error                                    { return nil }
 
-func TestParamsSetCorrectly(t *testing.T) {
+type hintRecordingQuerier struct {
+	storage.Querier
+
+	h *noopHintRecordingQueryable
+}
+
+func (h *hintRecordingQuerier) Select(sortSeries bool, hints *storage.SelectHints, matchers ...*labels.Matcher) storage.SeriesSet {
+	h.h.hints = append(h.h.hints, hints)
+	return h.Querier.Select(sortSeries, hints, matchers...)
+}
+
+func TestSelectHintsSetCorrectly(t *testing.T) {
 	opts := EngineOpts{
 		Logger:        nil,
 		Reg:           nil,
@@ -259,192 +252,139 @@ func TestParamsSetCorrectly(t *testing.T) {
 		LookbackDelta: 5 * time.Second,
 	}
 
-	cases := []struct {
+	for _, tc := range []struct {
 		query string
 
-		// All times are in seconds.
+		// All times are in milliseconds.
 		start int64
 		end   int64
 
-		paramStart int64
-		paramEnd   int64
-
-		paramGrouping []string
-		paramBy       bool
-		paramRange    int64
-		paramFunc     string
+		// TODO(bwplotka): Add support for better hints when subquerying.
+		expected []*storage.SelectHints
 	}{{
-		query: "foo",
-		start: 10,
-
-		paramStart: 5,
-		paramEnd:   10,
+		query: "foo", start: 10000,
+		expected: []*storage.SelectHints{
+			{Start: 5000, End: 10000},
+		},
 	}, {
-		query: "foo[2m]",
-		start: 200,
-
-		paramStart: 80, // 200 - 120
-		paramEnd:   200,
-		paramRange: 120000,
+		query: "foo[2m]", start: 200000,
+		expected: []*storage.SelectHints{
+			{Start: 80000, End: 200000, Range: 120000},
+		},
 	}, {
-		query: "foo[2m] offset 2m",
-		start: 300,
-
-		paramStart: 60,
-		paramEnd:   180,
-		paramRange: 120000,
+		query: "foo[2m] offset 2m", start: 300000,
+		expected: []*storage.SelectHints{
+			{Start: 60000, End: 180000, Range: 120000},
+		},
 	}, {
-		query: "foo[2m:1s]",
-		start: 300,
-
-		paramStart: 175, // 300 - 120 - 5
-		paramEnd:   300,
+		query: "foo[2m:1s]", start: 300000,
+		expected: []*storage.SelectHints{
+			{Start: 175000, End: 300000},
+		},
 	}, {
-		query: "count_over_time(foo[2m:1s])",
-		start: 300,
-
-		paramStart: 175, // 300 - 120 - 5
-		paramEnd:   300,
-		paramFunc:  "count_over_time",
+		query: "count_over_time(foo[2m:1s])", start: 300000,
+		expected: []*storage.SelectHints{
+			{Start: 175000, End: 300000, Func: "count_over_time"},
+		},
 	}, {
-		query: "count_over_time(foo[2m:1s] offset 10s)",
-		start: 300,
-
-		paramStart: 165, // 300 - 120 - 5 - 10
-		paramEnd:   300,
-		paramFunc:  "count_over_time",
+		query: "count_over_time(foo[2m:1s] offset 10s)", start: 300000,
+		expected: []*storage.SelectHints{
+			{Start: 165000, End: 290000, Func: "count_over_time"},
+		},
 	}, {
-		query: "count_over_time((foo offset 10s)[2m:1s] offset 10s)",
-		start: 300,
-
-		paramStart: 155, // 300 - 120 - 5 - 10 - 10
-		paramEnd:   290,
-		paramFunc:  "count_over_time",
+		query: "count_over_time((foo offset 10s)[2m:1s] offset 10s)", start: 300000,
+		expected: []*storage.SelectHints{
+			{Start: 155000, End: 280000, Func: "count_over_time"},
+		},
 	}, {
-		// Range queries now.
-		query: "foo",
-		start: 10,
-		end:   20,
 
-		paramStart: 5,
-		paramEnd:   20,
+		query: "foo", start: 10000, end: 20000,
+		expected: []*storage.SelectHints{
+			{Start: 5000, End: 20000, Step: 1000},
+		},
 	}, {
-		query: "rate(foo[2m])",
-		start: 200,
-		end:   500,
-
-		paramStart: 80, // 200 - 120
-		paramEnd:   500,
-		paramRange: 120000,
-		paramFunc:  "rate",
+		query: "rate(foo[2m])", start: 200000, end: 500000,
+		expected: []*storage.SelectHints{
+			{Start: 80000, End: 500000, Range: 120000, Func: "rate", Step: 1000},
+		},
 	}, {
-		query: "rate(foo[2m] offset 2m)",
-		start: 300,
-		end:   500,
-
-		paramStart: 60,
-		paramEnd:   380,
-		paramRange: 120000,
-		paramFunc:  "rate",
+		query: "rate(foo[2m] offset 2m)", start: 300000, end: 500000,
+		expected: []*storage.SelectHints{
+			{Start: 60000, End: 380000, Range: 120000, Func: "rate", Step: 1000},
+		},
 	}, {
-		query: "rate(foo[2m:1s])",
-		start: 300,
-		end:   500,
-
-		paramStart: 175, // 300 - 120 - 5
-		paramEnd:   500,
-		paramFunc:  "rate",
+		query: "rate(foo[2m:1s])", start: 300000, end: 500000,
+		expected: []*storage.SelectHints{
+			{Start: 175000, End: 500000, Func: "rate", Step: 1000},
+		},
 	}, {
-		query: "count_over_time(foo[2m:1s])",
-		start: 300,
-		end:   500,
-
-		paramStart: 175, // 300 - 120 - 5
-		paramEnd:   500,
-		paramFunc:  "count_over_time",
+		query: "count_over_time(foo[2m:1s])", start: 300000, end: 500000,
+		expected: []*storage.SelectHints{
+			{Start: 175000, End: 500000, Func: "count_over_time", Step: 1000},
+		},
 	}, {
-		query: "count_over_time(foo[2m:1s] offset 10s)",
-		start: 300,
-		end:   500,
-
-		paramStart: 165, // 300 - 120 - 5 - 10
-		paramEnd:   500,
-		paramFunc:  "count_over_time",
+		query: "count_over_time(foo[2m:1s] offset 10s)", start: 300000, end: 500000,
+		expected: []*storage.SelectHints{
+			{Start: 165000, End: 490000, Func: "count_over_time", Step: 1000},
+		},
 	}, {
-		query: "count_over_time((foo offset 10s)[2m:1s] offset 10s)",
-		start: 300,
-		end:   500,
-
-		paramStart: 155, // 300 - 120 - 5 - 10 - 10
-		paramEnd:   490,
-		paramFunc:  "count_over_time",
+		query: "count_over_time((foo offset 10s)[2m:1s] offset 10s)", start: 300000, end: 500000,
+		expected: []*storage.SelectHints{
+			{Start: 155000, End: 480000, Func: "count_over_time", Step: 1000},
+		},
 	}, {
-		query: "sum by (dim1) (foo)",
-		start: 10,
-
-		paramStart:    5,
-		paramEnd:      10,
-		paramGrouping: []string{"dim1"},
-		paramBy:       true,
-		paramFunc:     "sum",
+		query: "sum by (dim1) (foo)", start: 10000,
+		expected: []*storage.SelectHints{
+			{Start: 5000, End: 10000, Func: "sum", By: true, Grouping: []string{"dim1"}},
+		},
 	}, {
-		query: "sum without (dim1) (foo)",
-		start: 10,
-
-		paramStart:    5,
-		paramEnd:      10,
-		paramGrouping: []string{"dim1"},
-		paramBy:       false,
-		paramFunc:     "sum",
+		query: "sum without (dim1) (foo)", start: 10000,
+		expected: []*storage.SelectHints{
+			{Start: 5000, End: 10000, Func: "sum", Grouping: []string{"dim1"}},
+		},
 	}, {
-		query: "sum by (dim1) (avg_over_time(foo[1s]))",
-		start: 10,
-
-		paramStart:    9,
-		paramEnd:      10,
-		paramGrouping: nil,
-		paramBy:       false,
-		paramRange:    1000,
-		paramFunc:     "avg_over_time",
+		query: "sum by (dim1) (avg_over_time(foo[1s]))", start: 10000,
+		expected: []*storage.SelectHints{
+			{Start: 9000, End: 10000, Func: "avg_over_time", Range: 1000},
+		},
 	}, {
-		query: "sum by (dim1) (max by (dim2) (foo))",
-		start: 10,
-
-		paramStart:    5,
-		paramEnd:      10,
-		paramGrouping: []string{"dim2"},
-		paramBy:       true,
-		paramFunc:     "max",
+		query: "sum by (dim1) (max by (dim2) (foo))", start: 10000,
+		expected: []*storage.SelectHints{
+			{Start: 5000, End: 10000, Func: "max", By: true, Grouping: []string{"dim2"}},
+		},
 	}, {
-		query: "(max by (dim1) (foo))[5s:1s]",
-		start: 10,
+		query: "(max by (dim1) (foo))[5s:1s]", start: 10000,
+		expected: []*storage.SelectHints{
+			{Start: 0, End: 10000, Func: "max", By: true, Grouping: []string{"dim1"}},
+		},
+	}, {
+		query: "(sum(http_requests{group=~\"p.*\"})+max(http_requests{group=~\"c.*\"}))[20s:5s]", start: 120000,
+		expected: []*storage.SelectHints{
+			{Start: 95000, End: 120000, Func: "sum", By: true},
+			{Start: 95000, End: 120000, Func: "max", By: true},
+		},
+	}} {
+		t.Run(tc.query, func(t *testing.T) {
+			engine := NewEngine(opts)
+			hintsRecorder := &noopHintRecordingQueryable{}
 
-		paramStart:    0,
-		paramEnd:      10,
-		paramGrouping: []string{"dim1"},
-		paramBy:       true,
-		paramFunc:     "max",
-	}}
+			var (
+				query Query
+				err   error
+			)
+			if tc.end == 0 {
+				query, err = engine.NewInstantQuery(hintsRecorder, tc.query, timestamp.Time(tc.start))
+			} else {
+				query, err = engine.NewRangeQuery(hintsRecorder, tc.query, timestamp.Time(tc.start), timestamp.Time(tc.end), time.Second)
+			}
+			testutil.Ok(t, err)
 
-	for _, tc := range cases {
-		engine := NewEngine(opts)
-		queryable := storage.QueryableFunc(func(ctx context.Context, mint, maxt int64) (storage.Querier, error) {
-			return &hintCheckerQuerier{start: tc.paramStart * 1000, end: tc.paramEnd * 1000, grouping: tc.paramGrouping, by: tc.paramBy, selRange: tc.paramRange, function: tc.paramFunc, t: t}, nil
+			res := query.Exec(context.Background())
+			testutil.Ok(t, res.Err)
+
+			testutil.Equals(t, tc.expected, hintsRecorder.hints)
 		})
 
-		var (
-			query Query
-			err   error
-		)
-		if tc.end == 0 {
-			query, err = engine.NewInstantQuery(queryable, tc.query, time.Unix(tc.start, 0))
-		} else {
-			query, err = engine.NewRangeQuery(queryable, tc.query, time.Unix(tc.start, 0), time.Unix(tc.end, 0), time.Second)
-		}
-		testutil.Ok(t, err)
-
-		res := query.Exec(context.Background())
-		testutil.Ok(t, res.Err)
 	}
 }
 
@@ -499,9 +439,7 @@ func TestEngineShutdown(t *testing.T) {
 	testutil.NotOk(t, res2.Err, "expected error on querying with canceled context but got none")
 
 	var e ErrQueryCanceled
-	// TODO: when circleci-windows moves to go 1.13:
-	// testutil.Assert(t, errors.As(res2.Err, &e), "expected cancellation error but got: %s", res2.Err)
-	testutil.Assert(t, strings.HasPrefix(res2.Err.Error(), e.Error()), "expected cancellation error but got: %s", res2.Err)
+	testutil.Assert(t, errors.As(res2.Err, &e), "expected cancellation error but got: %s", res2.Err)
 }
 
 func TestEngineEvalStmtTimestamps(t *testing.T) {
@@ -599,9 +537,8 @@ load 10s
 		}
 
 		testutil.Ok(t, res.Err)
-		testutil.Equals(t, c.Result, res.Value)
+		testutil.Equals(t, c.Result, res.Value, "query %q failed", c.Query)
 	}
-
 }
 
 func TestMaxQuerySamples(t *testing.T) {
@@ -831,7 +768,7 @@ load 10s
 
 		res := qry.Exec(test.Context())
 		testutil.Equals(t, c.Result.Err, res.Err)
-		testutil.Equals(t, c.Result.Value, res.Value)
+		testutil.Equals(t, c.Result.Value, res.Value, "query %q failed", c.Query)
 	}
 }
 
@@ -839,7 +776,7 @@ func TestRecoverEvaluatorRuntime(t *testing.T) {
 	ev := &evaluator{logger: log.NewNopLogger()}
 
 	var err error
-	defer ev.recover(&err)
+	defer ev.recover(nil, &err)
 
 	// Cause a runtime panic.
 	var a []int
@@ -862,28 +799,50 @@ func TestRecoverEvaluatorError(t *testing.T) {
 			t.Fatalf("wrong error message: %q, expected %q", err, e)
 		}
 	}()
-	defer ev.recover(&err)
+	defer ev.recover(nil, &err)
+
+	panic(e)
+}
+
+func TestRecoverEvaluatorErrorWithWarnings(t *testing.T) {
+	ev := &evaluator{logger: log.NewNopLogger()}
+	var err error
+	var ws storage.Warnings
+
+	warnings := storage.Warnings{errors.New("custom warning")}
+	e := errWithWarnings{
+		err:      errors.New("custom error"),
+		warnings: warnings,
+	}
+
+	defer func() {
+		if err.Error() != e.Error() {
+			t.Fatalf("wrong error message: %q, expected %q", err, e)
+		}
+		if len(ws) != len(warnings) && ws[0] != warnings[0] {
+			t.Fatalf("wrong warning message: %q, expected %q", ws[0], warnings[0])
+		}
+	}()
+	defer ev.recover(&ws, &err)
 
 	panic(e)
 }
 
 func TestSubquerySelector(t *testing.T) {
-	tests := []struct {
+	type caseType struct {
+		Query  string
+		Result Result
+		Start  time.Time
+	}
+
+	for _, tst := range []struct {
 		loadString string
-		cases      []struct {
-			Query  string
-			Result Result
-			Start  time.Time
-		}
+		cases      []caseType
 	}{
 		{
 			loadString: `load 10s
 							metric 1 2`,
-			cases: []struct {
-				Query  string
-				Result Result
-				Start  time.Time
-			}{
+			cases: []caseType{
 				{
 					Query: "metric[20s:10s]",
 					Result: Result{
@@ -988,11 +947,7 @@ func TestSubquerySelector(t *testing.T) {
 							http_requests{job="api-server", instance="1", group="production"}	0+20x1000 200+30x1000
 							http_requests{job="api-server", instance="0", group="canary"}		0+30x1000 300+80x1000
 							http_requests{job="api-server", instance="1", group="canary"}		0+40x2000`,
-			cases: []struct {
-				Query  string
-				Result Result
-				Start  time.Time
-			}{
+			cases: []caseType{
 				{ // Normal selector.
 					Query: `http_requests{group=~"pro.*",instance="0"}[30s:10s]`,
 					Result: Result{
@@ -1093,32 +1048,27 @@ func TestSubquerySelector(t *testing.T) {
 				},
 			},
 		},
-	}
-
-	SetDefaultEvaluationInterval(1 * time.Minute)
-	for _, tst := range tests {
-		test, err := NewTest(t, tst.loadString)
-		testutil.Ok(t, err)
-
-		defer test.Close()
-
-		err = test.Run()
-		testutil.Ok(t, err)
-
-		engine := test.QueryEngine()
-		for _, c := range tst.cases {
-			var err error
-			var qry Query
-
-			qry, err = engine.NewInstantQuery(test.Queryable(), c.Query, c.Start)
+	} {
+		t.Run("", func(t *testing.T) {
+			test, err := NewTest(t, tst.loadString)
 			testutil.Ok(t, err)
+			defer test.Close()
 
-			res := qry.Exec(test.Context())
-			testutil.Equals(t, c.Result.Err, res.Err)
-			mat := res.Value.(Matrix)
-			sort.Sort(mat)
-			testutil.Equals(t, c.Result.Value, mat)
-		}
+			testutil.Ok(t, test.Run())
+			engine := test.QueryEngine()
+			for _, c := range tst.cases {
+				t.Run(c.Query, func(t *testing.T) {
+					qry, err := engine.NewInstantQuery(test.Queryable(), c.Query, c.Start)
+					testutil.Ok(t, err)
+
+					res := qry.Exec(test.Context())
+					testutil.Equals(t, c.Result.Err, res.Err)
+					mat := res.Value.(Matrix)
+					sort.Sort(mat)
+					testutil.Equals(t, c.Result.Value, mat)
+				})
+			}
+		})
 	}
 }
 
